{ **********************************************************************
  *                          Unit MATRICES.PAS                         *
  *                            Version 2.4d                            *
  *                   (c) J. Debord, September 2003                    *
  **********************************************************************
  This unit implements vector and matrix operations with dynamic arrays.
  There are 10 types available:

    TVector,     TMatrix      for floating point arrays
    TIntVector,  TIntMatrix   for integer arrays
    TBoolVector, TBoolMatrix  for boolean arrays

  To use these arrays in your programs, you must:

  (1) Declare variables of the appropriate type, e.g.

         var
           V : TVector;
           A : TMatrix;

  (2) Allocate each array BEFORE using it:

         DimVector(V, N);         creates vector V[0..N]
         DimMatrix(A, N, M);      creates matrix A[0..N, 0..M]
                                  where N, M are two integer variables

         If the allocation succeeds, all array elements are initialized
         to zero (for numeric arrays), False (for boolean arrays), or
         the null string (for string arrays). Otherwise, the array is
         initialized to NIL.

         The Dim... procedure may be used again to redimension the array.

  (3) Use arrays as in standard Pascal, noting that:

      (a) You cannot use the assignment operator (:=) to copy the
          contents of an array into another array. Writing B := A
          simply makes B point to the same memory block than A. You
          must use one of the provided Copy... procedures (see their
          documentation in the interface part of the unit).

      (b) All arrays begin at index 0, so that the 0-indexed element
          is always present, even if you don't use it.

      (c) A matrix is declared as an array of vectors, so that A[I]
          denotes the I-th vector of matrix A and may be used as any
          vector.

  (4) To deallocate an array, assign the value NIL

         V := nil;

  For more information, read the comments of each routine in the
  interface part of the unit, and check the demo programs.
  **********************************************************************
  References :
  1) 'Mathematiques et Statistiques' by H. Haut (PSI ed.) : GaussJordan
     and related functions
  2) EISPACK (http://www.netlib.org/eispack) : SV_Decomp
  3) 'Matrix Computations' by Golub & Van Loan : QR_Decomp & QR_Solve
     (Pascal implementation contributed by Mark Vaughan)
  ********************************************************************** }

unit matrices;

interface


{ **********************************************************************
  This section defines some error codes.
  ********************************************************************** }

const
  MACHEP = 1.192093E-7;               { Floating point precision: 2^(-23) }
  MAT_OK       =   0;  { No error }
  MAT_SINGUL   = - 1;  { Singular matrix }
  MAT_NON_CONV = - 2;  { Non convergence of iterative procedure }
  MAT_NOT_PD   = - 3;  { Matrix not positive definite }

{ **********************************************************************
  This section defines the vector and matrix types.
  ********************************************************************** }

type
  Float = Double;
  TVector     = array of Float;
  TIntVector  = array of Integer;
  TBoolVector = array of Boolean;

  TMatrix     = array of TVector;
  TIntMatrix  = array of TIntVector;
  TBoolMatrix = array of TBoolVector;

{ **********************************************************************
  This section defines a function of several variables
  ********************************************************************** }

type
  TFuncNVar = function(X : TVector) : Float;

{ ----------------------------------------------------------------------
  Min, max, sign and exchange
  ---------------------------------------------------------------------- }

procedure Swap(var X, Y : Float);   overload;   { Exchange 2 reals }
procedure Swap(var X, Y : Integer); overload;   { Exchange 2 integers }

{ **********************************************************************
  Memory allocation routines
  ********************************************************************** }

procedure DimVector(var V : TVector; Ubound : Integer); overload;
{ ----------------------------------------------------------------------
  Creates floating point vector V[0..Ubound]
  ---------------------------------------------------------------------- }

procedure DimVector(var V : TIntVector; Ubound : Integer); overload;
{ ----------------------------------------------------------------------
  Creates integer vector V[0..Ubound]
  ---------------------------------------------------------------------- }

procedure DimVector(var V : TBoolVector; Ubound : Integer); overload;
{ ----------------------------------------------------------------------
  Creates boolean vector V[0..Ubound]
  ---------------------------------------------------------------------- }

procedure DimMatrix(var A : TMatrix; Ubound1, Ubound2 : Integer); overload;
{ ----------------------------------------------------------------------
  Creates floating point matrix A[0..Ubound1, 0..Ubound2]
  ---------------------------------------------------------------------- }

procedure DimMatrix(var A : TIntMatrix; Ubound1, Ubound2 : Integer); overload;
{ ----------------------------------------------------------------------
  Creates integer matrix A[0..Ubound1, 0..Ubound2]
  ---------------------------------------------------------------------- }

procedure DimMatrix(var A : TBoolMatrix; Ubound1, Ubound2 : Integer); overload;
{ ----------------------------------------------------------------------
  Creates boolean matrix A[0..Ubound1, 0..Ubound2]
  ---------------------------------------------------------------------- }

{ **********************************************************************
  Routines for copying vectors and matrices
  ----------------------------------------------------------------------
  Lbound, Ubound   : indices of first and last vector elements
  Lbound1, Lbound2 : indices of first matrix element in each dimension
  Ubound1, Ubound2 : indices of last matrix element in each dimension
  ********************************************************************** }

procedure SwapRows(I, K : Integer; A : TMatrix; Lbound, Ubound : Integer);
{ ----------------------------------------------------------------------
  Exchanges rows I and K of matrix A
  ---------------------------------------------------------------------- }

procedure SwapCols(J, K : Integer; A : TMatrix; Lbound, Ubound : Integer);
{ ----------------------------------------------------------------------
  Exchanges columns J and K of matrix A
  ---------------------------------------------------------------------- }

procedure CopyVector(Dest, Source : TVector; Lbound, Ubound : Integer);
{ ----------------------------------------------------------------------
  Copies vector Source into vector Dest
  ---------------------------------------------------------------------- }

procedure CopyMatrix(Dest, Source : TMatrix;
                     Lbound1, Lbound2, Ubound1, Ubound2 : Integer);
{ ----------------------------------------------------------------------
  Copies matrix Source into matrix Dest
  ---------------------------------------------------------------------- }

procedure CopyRowFromVector(Dest : TMatrix; Source : TVector;
                            Lbound, Ubound, Row : Integer);
{ ----------------------------------------------------------------------
  Copies vector Source into line Row of matrix Dest
  ---------------------------------------------------------------------- }

procedure CopyColFromVector(Dest : TMatrix; Source : TVector;
                            Lbound, Ubound, Col : Integer);
{ ----------------------------------------------------------------------
  Copies vector Source into column Col of matrix Dest
  ---------------------------------------------------------------------- }

procedure CopyVectorFromRow(Dest : TVector; Source : TMatrix;
                            Lbound, Ubound, Row : Integer);
{ ----------------------------------------------------------------------
  Copies line Row of matrix Source into vector Dest
  ---------------------------------------------------------------------- }

procedure CopyVectorFromCol(Dest : TVector; Source : TMatrix;
                            Lbound, Ubound, Col : Integer);
{ ----------------------------------------------------------------------
  Copies column Col of matrix Source into vector Dest
  ---------------------------------------------------------------------- }

{ **********************************************************************
  Vector and matrix functions
  ********************************************************************** }

function Min(X : TVector; Lbound, Ubound : Integer) : Float; overload;
{ ----------------------------------------------------------------------
  Returns the lowest value of vector X
  ---------------------------------------------------------------------- }

function Min(X : TIntVector; Lbound, Ubound : Integer) : Integer; overload;
{ ----------------------------------------------------------------------
  Returns the lowest value of integer vector X
  ---------------------------------------------------------------------- }

function Max(X : TVector; Lbound, Ubound : Integer) : Float; overload;
{ ----------------------------------------------------------------------
  Returns the highest value of vector X
  ---------------------------------------------------------------------- }

function Max(X : TIntVector; Lbound, Ubound : Integer) : Integer; overload;
{ ----------------------------------------------------------------------
  Returns the highest value of integer vector X
  ---------------------------------------------------------------------- }

function GaussJordan(A              : TMatrix;
                     B              : TVector;
                     Lbound, Ubound : Integer;
                     A_inv          : TMatrix;
                     X              : TVector;
                     var Det        : Float) : Integer; overload;
{ ----------------------------------------------------------------------
  Solves the system A * X = B (where B and X are vectors)
  by the Gauss-Jordan method
  ----------------------------------------------------------------------
  Input parameters  : A      = system matrix
                      B      = constant vector
                      Lbound = index of first matrix element
                      Ubound = index of last matrix element
  ----------------------------------------------------------------------
  Output parameters : A_inv  = inverse matrix
                      X      = solution vector
                      Det    = determinant of A
  ----------------------------------------------------------------------
  Possible results  : MAT_OK
                      MAT_SINGUL
  ---------------------------------------------------------------------- }

function InvMat(A              : TMatrix;
                Lbound, Ubound : Integer;
                A_inv          : TMatrix) : Integer;
{ ----------------------------------------------------------------------
  Computes the inverse of a square matrix by the Gauss-Jordan method
  ----------------------------------------------------------------------
  Parameters : as in Gauss-Jordan
  ----------------------------------------------------------------------
  Possible results : MAT_OK
                     MAT_SINGUL
  ---------------------------------------------------------------------- }

function Det(A              : TMatrix;
             Lbound, Ubound : Integer) : Float;
{ ----------------------------------------------------------------------
  Computes the determinant of a square matrix by the Gauss-Jordan method
  ----------------------------------------------------------------------
  Parameters : as in Gauss-Jordan
  ----------------------------------------------------------------------
  Notes : (1) This procedure destroys the original matrix A
          (2) If the matrix is quasi-singular, the value 0 is returned
  ---------------------------------------------------------------------- }


implementation

  procedure DimVector(var V : TVector; Ubound : Integer); overload;
  var
    I : Integer;
  begin
    { Check bounds }
    if Ubound < 0 then begin
        V := nil;
        Exit;
      end;

    { Allocate vector }
    SetLength(V, Succ(Ubound));
    if V = nil then Exit;

    { Initialize vector }
    for I := 0 to Ubound do
      V[I] := 0.0;
  end;

  procedure DimVector(var V : TIntVector; Ubound : Integer); overload;
  var
    I : Integer;
  begin
    { Check bounds }
    if Ubound < 0 then begin
        V := nil;
        Exit;
      end;

    { Allocate vector }
    SetLength(V, Succ(Ubound));
    if V = nil then Exit;

    { Initialize vector }
    for I := 0 to Ubound do
        V[I] := 0;
  end;

  procedure DimVector(var V : TBoolVector; Ubound : Integer); overload;
  var
    I : Integer;
  begin
    { Check bounds }
    if Ubound < 0 then begin
        V := nil;
        Exit;
      end;

    { Allocate vector }
    SetLength(V, Succ(Ubound));
    if V = nil then Exit;

    { Initialize vector }
    for I := 0 to Ubound do
      V[I] := False;
  end;

  procedure DimMatrix(var A : TMatrix; Ubound1, Ubound2 : Integer); overload;
  var
    I, J : Integer;
  begin
    if (Ubound1 < 0) or (Ubound2 < 0) then begin
        A := nil;
        Exit;
      end;

    { Allocate matrix }
    SetLength(A, Succ(Ubound1), Succ(Ubound2));
    if A = nil then Exit;

    { Initialize matrix }
    for I := 0 to Ubound1 do
      for J := 0 to Ubound2 do
        A[I,J] := 0.0;
  end;

  procedure DimMatrix(var A : TIntMatrix; Ubound1, Ubound2 : Integer); overload;
  var
    I, J : Integer;
  begin
    { Check bounds }
    if (Ubound1 < 0) or (Ubound2 < 0) then begin
        A := nil;
        Exit;
      end;

    { Allocate matrix }
    SetLength(A, Succ(Ubound1), Succ(Ubound2));
    if A = nil then Exit;

    { Initialize matrix }
    for I := 0 to Ubound1 do
      for J := 0 to Ubound2 do
        A[I,J] := 0;
  end;

  procedure DimMatrix(var A : TBoolMatrix; Ubound1, Ubound2 : Integer); overload;
  var
    I, J : Integer;
  begin
    { Check bounds }
    if (Ubound1 < 0) or (Ubound2 < 0) then begin
        A := nil;
        Exit;
      end;

    { Allocate matrix }
    SetLength(A, Succ(Ubound1), Succ(Ubound2));
    if A = nil then Exit;

    { Initialize matrix }
    for I := 0 to Ubound1 do
      for J := 0 to Ubound2 do
        A[I,J] := False;
  end;

  procedure SwapRows(I, K : Integer; A : TMatrix; Lbound, Ubound : Integer);
  var
    J : Integer;
  begin
    for J := Lbound to Ubound do
      Swap(A[I,J], A[K,J]);
  end;

  procedure SwapCols(J, K : Integer; A : TMatrix; Lbound, Ubound : Integer);
  var
    I : Integer;
  begin
    for I := Lbound to Ubound do
      Swap(A[I,J], A[I,K]);
  end;

  procedure CopyVector(Dest, Source : TVector; Lbound, Ubound : Integer);
  var
    I : Integer;
  begin
    for I := Lbound to Ubound do
      Dest[I] := Source[I];
  end;

  procedure CopyMatrix(Dest, Source : TMatrix;
                       Lbound1, Lbound2, Ubound1, Ubound2 : Integer);
  var
    I, J : Integer;
  begin
    for I := Lbound1 to Ubound1 do
      for J := Lbound2 to Ubound2 do
        Dest[I,J] := Source[I,J];
  end;

  procedure CopyRowFromVector(Dest : TMatrix; Source : TVector;
                              Lbound, Ubound, Row : Integer);
  var
    J : Integer;
  begin
    for J := Lbound to Ubound do
      Dest[Row,J] := Source[J];
  end;

  procedure CopyColFromVector(Dest : TMatrix; Source : TVector;
                              Lbound, Ubound, Col : Integer);
  var
    I : Integer;
  begin
    for I := Lbound to Ubound do
      Dest[I,Col] := Source[I];
  end;

  procedure CopyVectorFromRow(Dest : TVector; Source : TMatrix;
                              Lbound, Ubound, Row : Integer);
  var
    J : Integer;
  begin
    for J := Lbound to Ubound do
      Dest[J] := Source[Row,J];
  end;

  procedure CopyVectorFromCol(Dest : TVector; Source : TMatrix;
                              Lbound, Ubound, Col : Integer);
  var
    I : Integer;
  begin
    for I := Lbound to Ubound do
      Dest[I] := Source[I,Col];
  end;

  function Min(X : TVector; Lbound, Ubound : Integer) : Float; overload;
  var
    Xmin : Float;
    I    : Integer;
  begin
    Xmin := X[Lbound];
    for I := Succ(Lbound) to Ubound do
      if X[I] < Xmin then Xmin := X[I];
    Min := Xmin;
  end;

  function Max(X : TVector; Lbound, Ubound : Integer) : Float; overload;
  var
    Xmax : Float;
    I    : Integer;
  begin
    Xmax := X[Lbound];
    for I := Succ(Lbound) to Ubound do
      if X[I] > Xmax then Xmax := X[I];
    Max := Xmax;
  end;

  function Min(X : TIntVector; Lbound, Ubound : Integer) : Integer; overload;
  var
    I, Xmin : Integer;
  begin
    Xmin := X[Lbound];
    for I := Succ(Lbound) to Ubound do
      if X[I] < Xmin then Xmin := X[I];
    Min := Xmin;
  end;

  function Max(X : TIntVector; Lbound, Ubound : Integer) : Integer; overload;
  var
    I, Xmax : Integer;
  begin
    Xmax := X[Lbound];
    for I := Succ(Lbound) to Ubound do
      if X[I] > Xmax then Xmax := X[I];
    Max := Xmax;
  end;

  function GaussJordan(A              : TMatrix;
                       B              : TVector;
                       Lbound, Ubound : Integer;
                       A_inv          : TMatrix;
                       X              : TVector;
                       var Det        : Float) : Integer; overload;
  var
    I, J, K    : Integer;     { Loop variables }
    Ik, Jk     : Integer;     { Pivot coordinates }
    Pvt        : Float;       { Pivot }
    T          : Float;       { Auxiliary variable }
    PRow, PCol : TIntVector;  { Store line and column of pivot }
    MCol       : TVector;     { Stores a column of the matrix }
  begin
    DimVector(PRow, Ubound);
    DimVector(PCol, Ubound);
    DimVector(MCol, Ubound);

    { Copy A into A_inv and B into X }
    CopyMatrix(A_inv, A, Lbound, Lbound, Ubound, Ubound);
    CopyVector(X, B, Lbound, Ubound);

    Det := 1.0;
    K := Lbound;
    while K <= Ubound do begin
        { Search for largest pivot in submatrix A_inv[K..Ubound, K..Ubound] }
        Pvt := A_inv[K,K];
        Ik := K;
        Jk := K;
        for I := K to Ubound do
          for J := K to Ubound do
            if Abs(A_inv[I,J]) > Abs(Pvt) then begin
                Pvt := A_inv[I,J];
                Ik := I;
                Jk := J;
              end;

        { Pivot too small ==> quasi-singular matrix }
        if Abs(Pvt) < MACHEP then begin
            GaussJordan := MAT_SINGUL;
            Exit;
          end;

        { Save pivot position }
        PRow[K] := Ik;
        PCol[K] := Jk;

        { Update determinant }
        Det := Det * Pvt;
        if Ik <> K then Det := - Det;
        if Jk <> K then Det := - Det;

        { Exchange current row (K) with pivot row (Ik) }
        if Ik <> K then begin
            SwapRows(Ik, K, A_inv, Lbound, Ubound);
            Swap(X[Ik], X[K]);
          end;

        { Exchange current column (K) with pivot column (Jk) }
        if Jk <> K then
          SwapCols(Jk, K, A_inv, Lbound, Ubound);

        { Store col. K of A_inv into MCol and set this col. to 0 }
        for I := Lbound to Ubound do
          if I <> K then begin
              MCol[I] := A_inv[I, K];
              A_inv[I, K] := 0.0;
            end
          else begin
              MCol[I] := 0.0;
              A_inv[I, K] := 1.0;
            end;

        { Transform pivot row }
        for J := Lbound to Ubound do
          A_inv[K,J] := A_inv[K,J] / Pvt;
        X[K] := X[K] / Pvt;

        { Transform other rows }
        for I := Lbound to Ubound do
          if I <> K then begin
              T := MCol[I];
              for J := Lbound to Ubound do
                A_inv[I,J] := A_inv[I,J] - T * A_inv[K,J];
              X[I] := X[I] - T * X[K];
            end;
        Inc(K);
      end;

    { Exchange rows of inverse matrix and solution vector }
    for I := Ubound downto Lbound do begin
        Ik := PCol[I];
        if Ik <> I then begin
            SwapRows(Ik, I, A_inv, Lbound, Ubound);
            Swap(X[Ik], X[I]);
          end;
      end;

    { Exchange columns of inverse matrix }
    for J := Ubound downto Lbound do begin
        Jk := PRow[J];
        if Jk <> J then
          SwapCols(Jk, J, A_inv, Lbound, Ubound);
      end;

    GaussJordan := MAT_OK;
  end;

  function InvMat(A              : TMatrix;
                  Lbound, Ubound : Integer;
                  A_inv          : TMatrix) : Integer;
  var
    I, J, K    : Integer;     { Loop variables }
    Ik, Jk     : Integer;     { Pivot coordinates }
    Pvt        : Float;       { Pivot }
    T          : Float;       { Auxiliary variable }
    PRow, PCol : TIntVector;  { Store line and column of pivot }
    MCol       : TVector;     { Stores a column of the matrix }
  begin
    DimVector(PRow, Ubound);
    DimVector(PCol, Ubound);
    DimVector(MCol, Ubound);

    { Copy A into A_inv }
    CopyMatrix(A_inv, A, Lbound, Lbound, Ubound, Ubound);

    K := Lbound;
    while K <= Ubound do begin
        { Search for largest pivot in submatrix A_inv[K..Ubound, K..Ubound] }
        Pvt := A_inv[K,K];
        Ik := K;
        Jk := K;
        for I := K to Ubound do
          for J := K to Ubound do
            if Abs(A_inv[I,J]) > Abs(Pvt) then begin
                Pvt := A_inv[I,J];
                Ik := I;
                Jk := J;
              end;

        { Pivot too small ==> quasi-singular matrix }
        if Abs(Pvt) < MACHEP then begin
            InvMat := MAT_SINGUL;
            Exit;
          end;

        { Save pivot position }
        PRow[K] := Ik;
        PCol[K] := Jk;

        { Exchange current row (K) with pivot row (Ik) }
        if Ik <> K then
          SwapRows(Ik, K, A_inv, Lbound, Ubound);

        { Exchange current column (K) with pivot column (Jk) }
        if Jk <> K then
          SwapCols(Jk, K, A_inv, Lbound, Ubound);

        { Store col. K of A_inv into MCol and set this col. to 0 }
        for I := Lbound to Ubound do
          if I <> K then begin
              MCol[I] := A_inv[I, K];
              A_inv[I, K] := 0.0;
            end
          else begin
              MCol[I] := 0.0;
              A_inv[I, K] := 1.0;
            end;

        { Transform pivot row }
        for J := Lbound to Ubound do
          A_inv[K,J] := A_inv[K,J] / Pvt;

        { Transform other rows }
        for I := Lbound to Ubound do
          if I <> K then begin
              T := MCol[I];
              for J := Lbound to Ubound do
                A_inv[I,J] := A_inv[I,J] - T * A_inv[K,J];
            end;
        Inc(K);
      end;

    { Exchange lines of inverse matrix }
    for I := Ubound downto Lbound do begin
        Ik := PCol[I];
        if Ik <> I then
          SwapRows(Ik, I, A_inv, Lbound, Ubound);
      end;

    { Exchange columns of inverse matrix }
    for J := Ubound downto Lbound do begin
        Jk := PRow[J];
        if Jk <> J then
          SwapCols(Jk, J, A_inv, Lbound, Ubound);
      end;

    InvMat := MAT_OK;
  end;

  function Det(A : TMatrix; Lbound, Ubound : Integer) : Float;
  var
    I, J, K    : Integer;     { Loop variables }
    Ik, Jk     : Integer;     { Pivot coordinates }
    Pvt        : Float;       { Pivot }
    T          : Float;       { Auxiliary variable }
    PRow, PCol : TIntVector;  { Store line and column of pivot }
    MCol       : TVector;     { Stores a column of the matrix }
    D          : Float;       { Determinant }
  begin
    DimVector(PRow, Ubound);
    DimVector(PCol, Ubound);
    DimVector(MCol, Ubound);

    D := 1.0;
    K := Lbound;
    while K <= Ubound do begin
        { Search for largest pivot in submatrix A[K..Ubound, K..Ubound] }
        Pvt := A[K,K];
        Ik := K;
        Jk := K;
        for I := K to Ubound do
          for J := K to Ubound do
            if Abs(A[I,J]) > Abs(Pvt) then begin
                Pvt := A[I,J];
                Ik := I;
                Jk := J;
              end;

        { Pivot too small ==> quasi-singular matrix }
        if Abs(Pvt) < MACHEP then begin
            Det := 0.0;
            Exit;
          end;

        { Save pivot position }
        PRow[K] := Ik;
        PCol[K] := Jk;

        { Update determinant }
        D := D * Pvt;
        if Ik <> K then D := - D;
        if Jk <> K then D := - D;

        { Exchange current row (K) with pivot row (Ik) }
        if Ik <> K then
          SwapRows(Ik, K, A, Lbound, Ubound);

        { Exchange current column (K) with pivot column (Jk) }
        if Jk <> K then
          SwapCols(Jk, K, A, Lbound, Ubound);

        { Store col. K of A into MCol and set this col. to 0 }
        for I := Lbound to Ubound do
          if I <> K then begin
              MCol[I] := A[I, K];
              A[I, K] := 0.0;
            end
          else begin
              MCol[I] := 0.0;
              A[I, K] := 1.0;
            end;

        { Transform pivot row }
        for J := Lbound to Ubound do
          A[K,J] := A[K,J] / Pvt;

        { Transform other rows }
        for I := Lbound to Ubound do
          if I <> K then begin
              T := MCol[I];
              for J := Lbound to Ubound do
                A[I,J] := A[I,J] - T * A[K,J];
            end;
        Inc(K);
      end;

    Det := D;
  end;

  procedure Swap(var X, Y : Integer); overload;
  var
    Temp : Integer;
  begin
    Temp := X;
    X := Y;
    Y := Temp;
  end;

  procedure Swap(var X, Y : Float); overload;
  var
    Temp : Float;
  begin
    Temp := X;
    X := Y;
    Y := Temp;
  end;

end.
